import pytest
from microscopy_portfolio import Portfolio
from microscopy_portfolio.portfolio_entry import PortfolioEntry

from .utils import (
    download_checker,
    portoflio_entry_checker,
    unique_md5_checker,
    unique_url_checker,
)

DATASETS = list(Portfolio().denoising)


@pytest.mark.dataset
@pytest.mark.parametrize("dataset", DATASETS)
def test_datasets(tmp_path, dataset: PortfolioEntry):
    """Test that all denoising datasets download properly.

    This test also checks the files and size.

    Parameters
    ----------
    tmp_path : Path
        Path to temporary directory.
    dataset : Dataset
        Dataset object.
    """
    download_checker(tmp_path, dataset)


def test_unique_hashes(portfolio: Portfolio):
    """Test that all denoising dataset hashes are unique.

    Parameters
    ----------
    portfolio : Portfolio
        Portfolio object.
    """
    unique_md5_checker(portfolio.denoising)


def test_unique_urls(portfolio: Portfolio):
    """Test that all denoising dataset URLs are unique.

    Parameters
    ----------
    portfolio : Portfolio
        Portfolio object.
    """
    unique_url_checker(portfolio.denoising)


def test_no_empty_dataset_entry(portfolio: Portfolio):
    """Test that no denoising dataset entry is empty.

    Parameters
    ----------
    portfolio : Portfolio
        Portfolio object.
    """
    for entry in portfolio.denoising:
        portoflio_entry_checker(entry)



import os
from pathlib import Path
from typing import Dict, List

import pytest
from microscopy_portfolio.portfolio import IterablePortfolio
from microscopy_portfolio.portfolio_entry import PortfolioEntry


def file_checker(path: Path, root_name: str, files: Dict[str, List[str]]) -> None:
    for folder, file_list in files.items():
        folder_path = path / root_name / folder
        for file in file_list:
            assert (
                folder_path / file
            ).is_file(), f"{file} does not exist in {folder_path}."


def unique_url_checker(iter_portfolio: IterablePortfolio) -> None:
    urls = []
    for entry in iter_portfolio:
        # add to list of urls
        urls.append(entry.url)

    assert len(urls) == len(set(urls)), f"Duplicated urls in {iter_portfolio.name}."


def unique_md5_checker(iter_portfolio: IterablePortfolio) -> None:
    hashes = []
    for entry in iter_portfolio:
        # add to list of hashes
        hashes.append(entry.md5_hash)

    assert len(hashes) == len(
        set(hashes)
    ), f"Duplicated hashes in {iter_portfolio.name}."


def portoflio_entry_checker(entry: PortfolioEntry) -> None:
    assert entry.name is not None and entry.name != "", f"Invalid name in {entry}"
    assert entry.url is not None and entry.url != "", f"Invalid url in {entry}"
    assert (
        entry.md5_hash is not None and entry.md5_hash != ""
    ), f"Invalid md5 hash in {entry}"
    assert (
        entry.description is not None and entry.description != ""
    ), f"Invalid description in {entry}"
    assert (
        entry.citation is not None and entry.citation != ""
    ), f"Invalid citation in {entry}"
    assert (
        entry.license is not None and entry.license != ""
    ), f"Invalid license in {entry}"
    assert (
        entry.file_name is not None and entry.file_name != ""
    ), f"Invalid file name in {entry}"
    assert entry.files is not None and len(entry.files) > 0, f"Invalid files in {entry}"
    assert entry.size is not None and entry.size > 0, f"Invalid size in {entry}"


def download_checker(path: Path, dataset: PortfolioEntry) -> None:
    # download dataset
    dataset.download(path, check_md5=True)

    # check that the zip file exists
    path_to_zip = path / dataset.file_name
    assert path_to_zip.exists(), f"{dataset.file_name} does not exist after download."

    # check that the files are there
    file_checker(path, dataset.file_name[:-4], dataset.files)

    # check file size with a tolerance of 5% or 3MB
    file_size = os.path.getsize(path_to_zip) / 1024 / 1024  # MB
    abs_tolerance = max(0.05 * dataset.size, 3)
    assert dataset.size == pytest.approx(file_size, abs=abs_tolerance), (
        f"{dataset.name} has not the expected size "
        f"(expected {dataset.size}, got {file_size})."
    )



import hashlib
import zipfile
from pathlib import Path
from typing import Dict, Union
from urllib import request

from tqdm import tqdm


# from https://stackoverflow.com/a/53877507
class DownloadProgressBar(tqdm):
    """Custom progress bar for download."""

    def update_to(
        self, b: float = 1, bsize: float = 1, tsize: Union[float, None] = None
    ) -> None:
        """Update tqdm progress bar.

        Parameters
        ----------
        b : float, optional
            Number of blocks transferred so far, by default 1
        bsize : float, optional
            Size of each block (in tqdm units), by default 1
        tsize : Union[float, None], optional
            Total size (in tqdm units), by default None
        """
        if tsize is not None:
            self.total = tsize
            self.update(b * bsize - self.n)


class PortfolioEntry:
    """Base class for portfolio entries.

    Attributes
    ----------
        name (str): Name of the dataset.
        url (str): URL of the dataset.
        description (str): Description of the dataset.
        license (str): License of the dataset.
        citation (str): Citation to use when referring to the dataset.
        file_name (str): Name of the downloaded file.
        md5_hash (str): MD5 hash of the downloaded file.
        files (dict[str, list]): Dictionary of files in the dataset.
        size (int): Size of the dataset in MB.
    """

    def __init__(
        self,
        name: str,
        url: str,
        description: str,
        license: str,
        citation: str,
        file_name: str,
        md5_hash: str,
        files: Dict[str, list],
        size: float = 0,
        **kwargs: str,
    ) -> None:
        self._name = name
        self._url = url
        self._description = description
        self._license = license
        self._citation = citation
        self._file_name = file_name
        self._md5_hash = md5_hash
        self._files = files
        self._size = size

    @property
    def name(self) -> str:
        """Name of the dataset.

        Returns
        -------
        str
            Name of the dataset.
        """
        return self._name

    @property
    def url(self) -> str:
        """URL of the dataset.

        Returns
        -------
        str
            URL of the dataset.
        """
        return self._url

    @property
    def description(self) -> str:
        """Description of the dataset.

        Returns
        -------
        str
            Description of the dataset.
        """
        return self._description

    @property
    def license(self) -> str:
        """License of the dataset.

        Returns
        -------
        str
            License of the dataset.
        """
        return self._license

    @property
    def citation(self) -> str:
        """Citation to use when referring to the dataset.

        Returns
        -------
        str
            Citation to use when referring to the dataset.
        """
        return self._citation

    @property
    def file_name(self) -> str:
        """Name of the downloaded file.

        Returns
        -------
        str
            Name of the downloaded file.
        """
        return self._file_name

    @property
    def md5_hash(self) -> str:
        """MD5 hash of the downloaded file.

        Returns
        -------
        str
            MD5 hash of the downloaded file.
        """
        return self._md5_hash

    @property
    def files(self) -> Dict[str, list]:
        """Dictionary of files in the dataset.

        Returns
        -------
        dict[str, list]
            Dictionary of files in the dataset.
        """
        return self._files

    @property
    def size(self) -> float:
        """Size of the dataset in MB.

        Returns
        -------
        float
            Size of the dataset in MB.
        """
        return self._size

    def download(
        self,
        path: Union[str, Path],
        check_md5: bool = True,
    ) -> dict:
        """Download dataset in the specified path.

        Parameters
        ----------
        path : str | Path
            Path to the folder in which to download the dataset.
        check_md5 : bool, optional
            Whether to check the MD5 hash of the downloaded file, by default True.

        Returns
        -------
        dict
            Dictionary of downloaded files.

        Raises
        ------
        ValueError
            If path is not a directory.
        ValueError
            If the md5 hash of the downloaded file is different from the expected one.
        """
        # TODO refactors in smaller functions
        path = Path(path)
        if path.exists() and not path.is_dir():
            raise ValueError(f"Path {path} is not a directory.")

        if not path.exists():
            path.mkdir()

        # check if zip file exists
        file_path = Path(path, self.file_name)
        if not file_path.exists():
            print(f"Downloading {self.name} to {path} might take some time.")

            # download data
            with DownloadProgressBar(
                unit="B", unit_scale=True, miniters=1, desc=self.url.split("/")[-1]
            ) as t:
                request.urlretrieve(
                    self.url, filename=file_path, reporthook=t.update_to
                )

            print("Download finished.")

        if not file_path.exists():
            raise ValueError(f"File {file_path} does not exist. Error downloading it.")

        # check if md5 hash is correct
        if check_md5:
            print(f"Checking MD5 hash of {file_path}.")

            # compute hash
            file_hash = hashlib.md5(open(file_path, "rb").read()).hexdigest()

            # compare with expected hash
            if file_hash != self.md5_hash:
                raise ValueError(
                    f"MD5 hash of {file_path} is not correct. "
                    f"Expected {self.md5_hash}, got {file_hash}."
                )

            print("MD5 hash is correct.")

        # unzip data
        data_path = Path(path, self.file_name[:-4])
        # TODO progress bar
        if zipfile.is_zipfile(file_path):
            print(f"Unzipping {file_path} to {data_path}.")

            # unzip data
            with zipfile.ZipFile(file_path, "r") as zip_ref:
                zip_ref.extractall(data_path)

            print("Unzipping finished.")

        return self.files

    def to_dict(self) -> dict:
        """Convert PortfolioEntry to a dictionnary.

        Returns
        -------
            dict: A dictionnary containing the PortfolioEntry attributes.
        """
        return {
            "name": self.name,
            "url": self.url,
            "description": self.description,
            "license": self.license,
            "citation": self.citation,
            "file_name": self.file_name,
            "md5_hash": self.md5_hash,
            "files": self.files,
            "size": self.size,
        }

    def __str__(self) -> str:
        """Convert PortfolioEntry to a string.

        Returns
        -------
        str: A string containing the PortfolioEntry attributes.
        """
        return str(self.to_dict())



import hashlib

import pytest
from microscopy_portfolio import Portfolio
from microscopy_portfolio.portfolio_entry import PortfolioEntry


class FaultyMD5(PortfolioEntry):
    """Faulty PortfolioEntry.

    A PortfolioEntry with a faulty md5 hash.

    Attributes
    ----------
    name (str): Name of the dataset.
    url (str): URL to the dataset.
    file_name (str): Name of the file.
    md5_hash (str): Faulty MD5 hash.
    description (str): Description of the dataset.
    citation (str): Citation of the dataset.
    license (str): License of the dataset.
    files (dict): Dictionary of files.
    """

    def __init__(self) -> None:
        super().__init__(
            name="Wikipedia logo",
            url="https://en.wikipedia.org/wiki/Wikipedia_logo#/media/File:Wikipedia-logo-v2.svg",
            file_name="Wikipedia-logo-v2.svg",
            md5_hash=hashlib.md5(b"I would prefer not to").hexdigest(),
            description="Wikipedia logo",
            citation="Wikipedia",
            license="CC BY-SA 3.0",
            files={
                ".": ["Wikipedia-logo-v2.svg"],
            },
            size=0.4,
        )


class PaleBlueDot(PortfolioEntry):
    """The original Pale Blue Dot image.

    Attributes
    ----------
    name (str): Name of the dataset.
    url (str): URL to the dataset.
    file_name (str): Name of the file.
    md5_hash (str): MD5 hash of the file.
    description (str): Description of the dataset.
    citation (str): Citation of the dataset.
    license (str): License of the dataset.
    files (dict): Dictionary of files.
    size (float): Size of the dataset in MB.
    """

    def __init__(self) -> None:
        super().__init__(
            name="Wikipedia logo",
            url="https://solarsystem.nasa.gov/rails/active_storage/blobs/eyJfcmFpbHMiOnsibWVzc2FnZSI6IkJBaHBBaUZoIiwiZXhwIjpudWxsLCJwdXIiOiJibG9iX2lkIn19--4b5b6d8ce74a6930534a08e4d7dd002f24f1efcb/P36254.jpg",
            file_name="P36254.jpg",
            md5_hash="42be58bb07b17df966f6ebc41941bac7",
            description="Pale Blue Dot, credit NASA/JPL-Caltech.",
            citation="NASA/JPL-Caltech",
            license="Unknown",
            files={
                ".": ["P36254.jpg"],
            },
            size=0.4,
        )


@pytest.fixture
def faulty_portfolio_entry() -> FaultyMD5:
    """Fixture for a faulty PortfolioEntry.

    Returns
    -------
    FaultyMD5
        A PortfolioEntry with a faulty md5 hash."""
    return FaultyMD5()


@pytest.fixture
def pale_blue_dot() -> PaleBlueDot:
    """Fixture for the PaleBlueDot.

    Returns
    -------
    PaleBlueDot
        The PaleBlueDot picture.
    """
    return PaleBlueDot()


@pytest.fixture
def portfolio() -> Portfolio:
    """Fixture for the Portfolio.

    Returns
    -------
    Portfolio
        The Portfolio.
    """
    return Portfolio()



from __future__ import annotations

import json
from dataclasses import dataclass
from json import JSONEncoder
from pathlib import Path

from .denoiseg_datasets import DSB2018, MouseNuclei, NoiseLevel, SegFlywing
from .denoising_datasets import N2V_BSD68, N2V_RGB, N2V_SEM, Convallaria, Flywing
from .portfolio_entry import PortfolioEntry


class IterablePortfolio:
    """Iterable portfolio class.

    Subclass this class and add PortfolioEntry objects as attributes.


    Attributes
    ----------
    _name : str
        Name of the portfolio.
    _datasets : List[PortfolioEntry]
        List of datasets in the portfolio.
    _current_index : int
    Current index of the iterator.
    """

    def __init__(self, name: str) -> None:
        self._name = name

        # create list of datasets
        datasets = []
        for dataset in vars(self).values():
            if isinstance(dataset, PortfolioEntry):
                datasets.append(dataset)

        # record datasets
        self._datasets = datasets
        self._current_index = 0

    def __iter__(self) -> IterablePortfolio:
        """Iterator method.

        Returns
        -------
        IterablePortfolio
            Iterator over the portfolio.
        """
        return self

    def __next__(self) -> PortfolioEntry:
        """Next method.

        Returns
        -------
        PortfolioEntry
            Next dataset in the portfolio.
        """
        if self._current_index < len(self._datasets):
            next_dataset = self._datasets[self._current_index]
            self._current_index += 1
            return next_dataset
        raise StopIteration("The iterator does not have any more elements.")

    @property
    def name(self) -> str:
        """Name of the portfolio.

        Returns
        -------
        str
            Name of the portfolio.
        """
        return self._name

    def list_datasets(self) -> list[str]:
        """List datasets in the portfolio using friendly names.

        The friendly names are the names of the portfolio entries, rather
        than that of the IterablePortfolio attributes.

        Returns
        -------
        list[str]
            List of datasets in the portfolio.
        """
        attributes = []

        # for each attribute
        for attribute in vars(self).values():
            if isinstance(attribute, PortfolioEntry):
                attributes.append(attribute.name)

        return attributes

    def as_dict(self) -> dict[str, dict[str, str]]:
        """Dictionary representation of a portfolio.

        Used to serialize the class to json, with friendly name as entries.

        Returns
        -------
        dict[str]
            Dictionary representation of the DenoiSeg portfolio.
        """
        entries = {}

        # for each attribute
        for attribute in vars(self).values():
            # if the attribute is a PortfolioEntry
            if isinstance(attribute, PortfolioEntry):
                # add the attribute to the entries dictionary
                entries[attribute.name] = {
                    "URL": attribute.url,
                    "Description": attribute.description,
                    "Citation": attribute.citation,
                    "License": attribute.license,
                    "File size": f"{attribute.size} MB",
                }
        return entries

    def __str__(self) -> str:
        """String representation of a portfolio.

        Returns
        -------
        str
        String representation of a portfolio.
        """
        return f"{self.name} datasets: {self.list_datasets()}"


class ItarablePortfolioEncoder(JSONEncoder):
    """Portfolio encoder class."""

    def default(self, o: IterablePortfolio) -> dict[str, dict[str, str]]:
        """Default method for json export.

        Parameters
        ----------
        o : IterablePortfolio
            Portfolio to export.

        Returns
        -------
        dict[str, str]
            Dictionary representation of the portfolio.
        """
        return o.as_dict()


class DenoiSeg(IterablePortfolio):
    """An IterablePortfolio of DenoiSeg datasets.

    Attributes
    ----------
        DSB2018_n0 (DSB2018): DSB2018 dataset with noise level 0.
        DSB2018_n10 (DSB2018): DSB2018 dataset with noise level 10.
        DSB2018_n20 (DSB2018): DSB2018 dataset with noise level 20.
        Flywing_n0 (SegFlywing): Flywing dataset with noise level 0.
        Flywing_n10 (SegFlywing): Flywing dataset with noise level 10.
        Flywing_n20 (SegFlywing): Flywing dataset with noise level 20.
        MouseNuclei_n0 (MouseNuclei): MouseNuclei dataset with noise level 0.
        MouseNuclei_n10 (MouseNuclei): MouseNuclei dataset with noise level 10.
        MouseNuclei_n20 (MouseNuclei): MouseNuclei dataset with noise level 20.
    """

    def __init__(self) -> None:
        self._DSB2018_n0 = DSB2018(NoiseLevel.N0)
        self._DSB2018_n10 = DSB2018(NoiseLevel.N10)
        self._DSB2018_n20 = DSB2018(NoiseLevel.N20)
        self._SegFlywing_n0 = SegFlywing(NoiseLevel.N0)
        self._SegFlywing_n10 = SegFlywing(NoiseLevel.N10)
        self._SegFlywing_n20 = SegFlywing(NoiseLevel.N20)
        self._MouseNuclei_n0 = MouseNuclei(NoiseLevel.N0)
        self._MouseNuclei_n10 = MouseNuclei(NoiseLevel.N10)
        self._MouseNuclei_n20 = MouseNuclei(NoiseLevel.N20)

        super().__init__("DenoiSeg")

    @property
    def DSB2018_n0(self) -> DSB2018:
        """DSB2018 dataset with noise level 0.

        Returns
        -------
        DSB2018
            DSB2018 dataset with noise level 0.
        """
        return self._DSB2018_n0

    @property
    def DSB2018_n10(self) -> DSB2018:
        """DSB2018 dataset with noise level 10.

        Returns
        -------
        DSB2018
            DSB2018 dataset with noise level 10.
        """
        return self._DSB2018_n10

    @property
    def DSB2018_n20(self) -> DSB2018:
        """DSB2018 dataset with noise level 20.

        Returns
        -------
        DSB2018
            DSB2018 dataset with noise level 20.
        """
        return self._DSB2018_n20

    @property
    def Flywing_n0(self) -> SegFlywing:
        """Flywing dataset with noise level 0.

        Returns
        -------
        SegFlywing
            Flywing dataset with noise level 0.
        """
        return self._SegFlywing_n0

    @property
    def Flywing_n10(self) -> SegFlywing:
        """Flywing dataset with noise level 10.

        Returns
        -------
        SegFlywing
            Flywing dataset with noise level 10.
        """
        return self._SegFlywing_n10

    @property
    def Flywing_n20(self) -> SegFlywing:
        """Flywing dataset with noise level 20.

        Returns
        -------
        SegFlywing
            Flywing dataset with noise level 20.
        """
        return self._SegFlywing_n20

    @property
    def MouseNuclei_n0(self) -> MouseNuclei:
        """MouseNuclei dataset with noise level 0.

        Returns
        -------
        MouseNuclei
            MouseNuclei dataset with noise level 0.
        """
        return self._MouseNuclei_n0

    @property
    def MouseNuclei_n10(self) -> MouseNuclei:
        """MouseNuclei dataset with noise level 10.

        Returns
        -------
        MouseNuclei
            MouseNuclei dataset with noise level 10.
        """
        return self._MouseNuclei_n10

    @property
    def MouseNuclei_n20(self) -> MouseNuclei:
        """MouseNuclei dataset with noise level 20.

        Returns
        -------
        MouseNuclei
            MouseNuclei dataset with noise level 20.
        """
        return self._MouseNuclei_n20


class Denoising(IterablePortfolio):
    """An IterablePortfolio of denoising datasets.

    Attributes
    ----------
    N2V_BSD68 (N2V_BSD68): BSD68 dataset.
    N2V_SEM (N2V_SEM): SEM dataset.
    N2V_RGB (N2V_RGB): RGB dataset.
    flywing (Flywing): Flywing dataset.
    Convallaria (Convallaria): Convallaria dataset.
    """

    def __init__(self) -> None:
        self._N2V_BSD68 = N2V_BSD68()
        self._N2V_SEM = N2V_SEM()
        self._N2V_RGB = N2V_RGB()
        self._flywing = Flywing()
        self._Convallaria = Convallaria()

        super().__init__("Denoising")

    @property
    def N2V_BSD68(self) -> N2V_BSD68:
        """BSD68 dataset.

        Returns
        -------
        N2V_BSD68
            BSD68 dataset.
        """
        return self._N2V_BSD68

    @property
    def N2V_SEM(self) -> N2V_SEM:
        """SEM dataset.

        Returns
        -------
        N2V_SEM
            SEM dataset.
        """
        return self._N2V_SEM

    @property
    def N2V_RGB(self) -> N2V_RGB:
        """RGB dataset.

        Returns
        -------
        N2V_RGB
            RGB dataset.
        """
        return self._N2V_RGB

    @property
    def flywing(self) -> Flywing:
        """Flywing dataset.

        Returns
        -------
        Flywing
            Flywing dataset.
        """
        return self._flywing

    @property
    def Convallaria(self) -> Convallaria:
        """Convallaria dataset.

        Returns
        -------
        Convallaria
            Convallaria dataset.
        """
        return self._Convallaria


@dataclass
class Portfolio:
    """Portfolio of datasets.

    Attributes
    ----------
    denoising (Denoising): Denoising datasets.
    denoiseg (DenoiSeg): DenoiSeg datasets.
    """

    def __init__(self) -> None:
        self._denoising = Denoising()
        self._denoiseg = DenoiSeg()

    @property
    def denoising(self) -> Denoising:
        """Denoising datasets.

        Returns
        -------
        Denoising
        Denoising datasets.
        """
        return self._denoising

    @property
    def denoiseg(self) -> DenoiSeg:
        """DenoiSeg datasets.

        Returns
        -------
        DenoiSeg
        DenoiSeg datasets.
        """
        return self._denoiseg

    def __str__(self) -> str:
        """String representation of the portfolio.

        This method allows having a frendly representation of the portfolio as string.

        Returns
        -------
        str
        String representation of the portfolio.
        """
        return (
            f"Portfolio:\n"
            f"Denoising datasets: {self.denoising.list_datasets()}\n"
            f"DenoiSeg datasets: {self.denoiseg.list_datasets()}"
        )

    def as_dict(self) -> dict[str, IterablePortfolio]:
        """Portfolio as dictionary.

        This method is used during json serialization to maintain human readable
        keys.

        Returns
        -------
        dict[str, IterablePortfolio]
            Portfolio as dictionary.
        """
        attributes = {}

        for attribute in vars(self).values():
            attributes[attribute.name] = attribute

        return attributes

    def to_json(self, path: str | Path) -> None:
        """Save portfolio to json file using the `as_dict` method.

        Parameters
        ----------
        path : str or Path
            Path to json file.
        """
        with open(path, "w") as f:
            json.dump(self.as_dict(), f, indent=4, cls=ItarablePortfolioEncoder)



"""microscopy_protfolio.

A helper package to download example datasets used in various publications and
deep-learning algorithms, including data featured in N2V, P(P)N2V, DivNoising, HDN,
EmbedSeg, etc.
"""
__all__ = ["Portfolio", "__author__", "__email__", "__version__"]

from importlib.metadata import PackageNotFoundError, version

from .portfolio import Portfolio

try:
    __version__ = version("microscopy-portfolio")
except PackageNotFoundError:
    __version__ = "uninstalled"
__author__ = "Joran Deschamps"
__email__ = "joran.deschamps@fht.org"



import pytest
from microscopy_portfolio.portfolio_entry import PortfolioEntry


def test_download(tmp_path, pale_blue_dot: PortfolioEntry):
    pale_blue_dot.download(tmp_path)


def test_download_in_invalid_path(tmp_path, pale_blue_dot: PortfolioEntry):
    """Test that downloading to an invalid path raises an error."""
    file_name = "file.txt"
    with open(tmp_path / file_name, "w") as f:
        f.write("CATS ARE NICE.")

    with pytest.raises(ValueError):
        pale_blue_dot.download(tmp_path / file_name)


def test_faulty_md5(tmp_path, faulty_portfolio_entry: PortfolioEntry):
    """Test that faulty md5 hash raises an error.

    Parameters
    ----------
        tmp_path (Path): Path to temporary directory.
        faulty_portfolio_entry (FaultyMD5): test PortfolioEntry
    """

    # test that it downloads fine when not checking md5 hash
    faulty_portfolio_entry.download(tmp_path, check_md5=False)
    assert (tmp_path / faulty_portfolio_entry.file_name).exists()

    with pytest.raises(ValueError):
        faulty_portfolio_entry.download(tmp_path, check_md5=True)


def test_change_entry(faulty_portfolio_entry):
    """Check that changing a PortfolioEntry member raises an error.

    Parameters
    ----------
    faulty_portfolio_entry : FaultyMD5
        Test PortfolioEntry.
    """
    # Verify that we can access the members
    faulty_portfolio_entry.name
    faulty_portfolio_entry.url
    faulty_portfolio_entry.description
    faulty_portfolio_entry.license
    faulty_portfolio_entry.citation
    faulty_portfolio_entry.file_name
    faulty_portfolio_entry.md5_hash
    faulty_portfolio_entry.files

    # Check that changing members raises errors
    with pytest.raises(AttributeError):
        faulty_portfolio_entry.name = ""

    with pytest.raises(AttributeError):
        faulty_portfolio_entry.url = ""

    with pytest.raises(AttributeError):
        faulty_portfolio_entry.description = ""

    with pytest.raises(AttributeError):
        faulty_portfolio_entry.license = ""

    with pytest.raises(AttributeError):
        faulty_portfolio_entry.citation = ""

    with pytest.raises(AttributeError):
        faulty_portfolio_entry.file_name = ""

    with pytest.raises(AttributeError):
        faulty_portfolio_entry.md5_hash = ""

    with pytest.raises(AttributeError):
        faulty_portfolio_entry.files = {}



from .portfolio_entry import PortfolioEntry

# TODO add descriptions in the class docstrings


class N2V_BSD68(PortfolioEntry):
    """BSD68 dataset.

    Attributes
    ----------
    name (str): Name of the dataset.
    url (str): URL of the dataset.
    file_name (str): Name of the downloaded file.
    md5_hash (str): MD5 hash of the downloaded file.
    description (str): Description of the dataset.
    license (str): License of the dataset.
    citation (str): Citation to use when referring to the dataset.
    files (dict): Dictionary containing the files to download.
    size (float): Size of the dataset in MB.
    """

    def __init__(self) -> None:
        super().__init__(
            name="N2V_BSD68",
            url="https://download.fht.org/jug/n2v/BSD68_reproducibility.zip",
            file_name="BSD68_reproducibility.zip",
            md5_hash="292c29895fa56ef7226487005b5955a2",
            description="This dataset is taken from K. Zhang et al (TIP, 2017). \n"
            "It consists of 400 gray-scale 180x180 images (cropped from the "
            "BSD dataset) and splitted between training and validation, and "
            "68 gray-scale test images (BSD68).\n"
            "All images were corrupted with Gaussian noise with standard "
            "deviation of 25 pixels. The test dataset contains the uncorrupted "
            "images as well.\n"
            "Original dataset: "
            "https://www2.eecs.berkeley.edu/Research/Projects/CS/vision/bsds/",
            license="Unknown",
            citation='D. Martin, C. Fowlkes, D. Tal and J. Malik, "A database of '
            "human segmented natural images and its application to "
            "evaluating segmentation algorithms and measuring ecological "
            'statistics," Proceedings Eighth IEEE International '
            "Conference on Computer Vision. ICCV 2001, Vancouver, BC, "
            "Canada, 2001, pp. 416-423 vol.2, doi: "
            "10.1109/ICCV.2001.937655.",
            files={
                "BSD68_reproducibility_data/test": [
                    "bsd68_gaussian25.npy",
                    "bsd68_groundtruth.npy",
                ],
                "BSD68_reproducibility_data/train": ["DCNN400_train_gaussian25.npy"],
                "BSD68_reproducibility_data/val": ["DCNN400_validation_gaussian25.npy"],
            },
            size=875.0,
        )


class N2V_SEM(PortfolioEntry):
    """SEM dataset.

    Attributes
    ----------
    name (str): Name of the dataset.
    url (str): URL of the dataset.
    file_name (str): Name of the downloaded file.
    md5_hash (str): MD5 hash of the downloaded file.
    description (str): Description of the dataset.
    license (str): License of the dataset.
    citation (str): Citation to use when referring to the dataset.
    files (dict): Dictionary containing the files to download.
    size (float): Size of the dataset in MB.
    """

    def __init__(self) -> None:
        super().__init__(
            name="N2V_SEM",
            url="https://download.fht.org/jug/n2v/SEM.zip",
            file_name="SEM.zip",
            md5_hash="953a815333805a423b7019bd16cc3341",
            description="Cropped images from a SEM dataset from T.-O. Buchholz et al "
            "(Methods Cell Biol, 2020).",
            license="CC-BY-4.0",
            citation="T.-O. Buchholz, A. Krull, R. Shahidi, G. Pigino, G. Jékely, "
            'F. Jug, "Content-aware image restoration for electron '
            'microscopy", Methods Cell Biol 152, 277-289',
            files={
                ".": ["train.tif", "validation.tif"],
            },
            size=13.0,
        )


class N2V_RGB(PortfolioEntry):
    """RGB dataset.

    Attributes
    ----------
    name (str): Name of the dataset.
    url (str): URL of the dataset.
    file_name (str): Name of the downloaded file.
    md5_hash (str): MD5 hash of the downloaded file.
    description (str): Description of the dataset.
    license (str): License of the dataset.
    citation (str): Citation to use when referring to the dataset.
    files (dict): Dictionary containing the files to download.
    size (float): Size of the dataset in MB.
    """

    def __init__(self) -> None:
        super().__init__(
            name="N2V_RGB",
            url="https://download.fht.org/jug/n2v/RGB.zip",
            file_name="RGB.zip",
            md5_hash="ad80d2fee3ae0a93208687e30ad2b63a",
            description="Banner of the CVPR 2019 conference with extra noise.",
            license="CC-BY-4.0",
            citation='A. Krull, T.-O. Buchholz and F. Jug, "Noise2Void - Learning '
            'Denoising From Single Noisy Images," 2019 IEEE/CVF '
            "Conference on Computer Vision and Pattern Recognition (CVPR),"
            " 2019, pp. 2124-2132",
            files={
                ".": ["longBeach.png"],
            },
            size=10.4,
        )


class Flywing(PortfolioEntry):
    """Flywing dataset.

    Attributes
    ----------
    name (str): Name of the dataset.
    url (str): URL of the dataset.
    file_name (str): Name of the downloaded file.
    md5_hash (str): MD5 hash of the downloaded file.
    description (str): Description of the dataset.
    license (str): License of the dataset.
    citation (str): Citation to use when referring to the dataset.
    files (dict): Dictionary containing the files to download.
    size (float): Size of the dataset in MB.
    """

    def __init__(self) -> None:
        super().__init__(
            name="Flywing",
            url="https://download.fht.org/jug/n2v/flywing-data.zip",
            file_name="flywing-data.zip",
            md5_hash="769f4e265f8ab8ccea1893087df019da",
            description="Image of a membrane-labeled fly wing (35x692x520 pixels).",
            license="CC-BY-4.0",
            citation="Buchholz, T.O., Prakash, M., Schmidt, D., Krull, A., Jug, "
            "F.: Denoiseg: joint denoising and segmentation. In: European "
            "Conference on Computer Vision (ECCV). pp. 324-337. Springer (2020) 8, 9",
            files={
                ".": ["flywing.tif"],
            },
            size=10.2,
        )


class Convallaria(PortfolioEntry):
    """Convallaria dataset.

    Attributes
    ----------
    name (str): Name of the dataset.
    url (str): URL of the dataset.
    file_name (str): Name of the file to download.
    md5_hash (str): MD5 hash of the file to download.
    description (str): Description of the dataset.
    license (str): License of the dataset.
    citation (str): Citation to use when referring to the dataset.
    files (dict): Dictionary containing the files to download.
    size (float): Size of the dataset in MB.
    """

    def __init__(self) -> None:
        super().__init__(
            name="Convallaria",
            url="https://cloud.mpi-cbg.de/index.php/s/BE8raMtHQlgLDF3/download",
            file_name="Convallaria_diaphragm.zip",
            md5_hash="7b8df3a83939decaede6753b8d38b52f",
            description="Image of a convallaria flower (35x692x520 pixels).\n"
            "The image also comes with a defocused image in order to allow \n"
            "estimating the noise distribution.",
            license="CC-BY-4.0",
            citation="Krull, A., Vičar, T., Prakash, M., Lalit, M., & Jug, F. (2020). "
            "Probabilistic noise2void: Unsupervised content-aware denoising. Frontiers"
            " in Computer Science, 2, 5.",
            files={
                "Convallaria_diaphragm": [
                    "20190520_tl_25um_50msec_05pc_488_130EM_Conv.tif",
                    "20190726_tl_50um_500msec_wf_130EM_FD.tif",
                ],
            },
            size=344.0,
        )



from enum import Enum

from .portfolio_entry import PortfolioEntry


class NoiseLevel(str, Enum):
    """An IntEnum representing the noise level of a dataset.

    N0 corresponds to the noise-free version of the dataset, N10 and N20 to
    images corrupted with Gaussian noise with zero-mean and standard deviations
    of 10 and 20, respectively.
    """

    N0 = "0"
    N10 = "10"
    N20 = "20"


class NoisyObject:
    """A mixin class for datasets with different noise levels.

    Attributes
    ----------
    noise_level (NoiseLevel): Noise level of the dataset.
    """

    def __init__(self, noise_level: NoiseLevel = NoiseLevel.N0, **kwargs: str) -> None:
        self._noise_level = noise_level

    @property
    def noise_level(self) -> NoiseLevel:
        """Noise level of the dataset."""
        return self._noise_level


class DSB2018(PortfolioEntry, NoisyObject):
    """The 2018 Data Science Bowl dataset used by DenoiSeg.

    The dataset is available in three different noise levels: N0, N10 and N20.


    Attributes
    ----------
    noise_level (NoiseLevel): Noise level of the dataset.
    """

    def __init__(self, noise_level: NoiseLevel = NoiseLevel.N0) -> None:
        """Initialize a DSB2018 instance.

        Parameters
        ----------
        noise_level : NoiseLevel, optional
            Noise level of the dataset, by default NoiseLevel.N0
        """
        super().__init__(
            noise_level=noise_level,
            name=f"DSB2018_n{noise_level.value}",
            url=self._get_url(noise_level),
            file_name=f"DSB2018_n{noise_level.value}.zip",
            md5_hash=self._get_hash(noise_level),
            description="From the Kaggle 2018 Data Science Bowl challenge, the "
            "training and validation sets consist of 3800 and 670 patches "
            "respectively, while the test set counts 50 images.\n"
            "Original data: "
            "https://www.kaggle.com/competitions/data-science-bowl-2018/data",
            license="GPL-3.0",
            citation="Caicedo, J.C., Goodman, A., Karhohs, K.W. et al. Nucleus "
            "segmentation across imaging experiments: the 2018 Data Science "
            "Bowl. Nat Methods 16, 1247-1253 (2019). "
            "https://doi.org/10.1038/s41592-019-0612-7",
            files={
                f"DSB2018_n{noise_level.value}/train": ["train_data.npz"],
                f"DSB2018_n{noise_level.value}/test": ["test_data.npz"],
            },
            size=self._get_size(noise_level),
        )

    @staticmethod
    def _get_url(noise: NoiseLevel) -> str:
        if noise == NoiseLevel.N0:
            return "https://zenodo.org/record/5156969/files/DSB2018_n0.zip?download=1"
        elif noise == NoiseLevel.N10:
            return "https://zenodo.org/record/5156977/files/DSB2018_n10.zip?download=1"
        else:
            return "https://zenodo.org/record/5156983/files/DSB2018_n20.zip?download=1"

    @staticmethod
    def _get_hash(noise: NoiseLevel) -> str:
        if noise == NoiseLevel.N0:
            return "80513b1eda8e08df1d8dcc5543ad1ad1"
        elif noise == NoiseLevel.N10:
            return "aa16c116949d8b8cd573d7bbeacbd0c3"
        else:
            return "81abc17313582a4f04f501e3dce1fe88"

    @staticmethod
    def _get_size(noise: NoiseLevel) -> float:
        if noise == NoiseLevel.N0:
            return 40.2
        elif noise == NoiseLevel.N10:
            return 366.0
        else:
            return 368.0


class SegFlywing(PortfolioEntry, NoisyObject):
    """Flywing dataset used by DenoiSeg.

    The dataset is available in three different noise levels: N0, N10 and N20.


    Attributes
    ----------
    noise_level (NoiseLevel): Noise level of the dataset.
    """

    def __init__(self, noise_level: NoiseLevel = NoiseLevel.N0) -> None:
        """Initialize a Flywing instance.

        Parameters
        ----------
        noise_level : NoiseLevel, optional
            Noise level of the dataset, by default NoiseLevel.N0
        """
        super().__init__(
            noise_level=noise_level,
            name=f"Flywing_n{noise_level.value}",
            url=self._get_url(noise_level),
            file_name=f"Flywing_n{noise_level.value}.zip",
            md5_hash=self._get_hash(noise_level),
            description="This dataset consist of 1428 training and 252 "
            "validation patches of a membrane labeled fly wing. The test set "
            "is comprised of 50 additional images.",
            license="CC BY-SA 4.0",
            citation="Buchholz, T.O., Prakash, M., Schmidt, D., Krull, A., Jug, "
            "F.: Denoiseg: joint denoising and segmentation. In: European "
            "Conference on Computer Vision (ECCV). pp. 324-337. Springer (2020) 8, 9",
            files={
                f"Flywing_n{noise_level.value}/train": ["train_data.npz"],
                f"Flywing_n{noise_level.value}/test": ["test_data.npz"],
            },
            size=self._get_size(noise_level),
        )

    @staticmethod
    def _get_url(noise: NoiseLevel) -> str:
        if noise == NoiseLevel.N0:
            return "https://zenodo.org/record/5156991/files/Flywing_n0.zip?download=1"
        elif noise == NoiseLevel.N10:
            return "https://zenodo.org/record/5156993/files/Flywing_n10.zip?download=1"
        else:
            return "https://zenodo.org/record/5156995/files/Flywing_n20.zip?download=1"

    @staticmethod
    def _get_hash(noise: NoiseLevel) -> str:
        if noise == NoiseLevel.N0:
            return "09e0af44f0f9862abae3816d7069604a"
        elif noise == NoiseLevel.N10:
            return "64d5300073e02c9651ec88c368c302e8"
        else:
            return "b8fbb96026bd10fd034b8c1270f6edbb"

    @staticmethod
    def _get_size(noise: NoiseLevel) -> float:
        if noise == NoiseLevel.N0:
            return 47.0
        elif noise == NoiseLevel.N10:
            return 282.0
        else:
            return 293.0


class MouseNuclei(PortfolioEntry, NoisyObject):
    """Mouse nuclei dataset used by DenoiSeg.

    The dataset is available in three different noise levels: N0, N10 and N20.


    Attributes
    ----------
    noise_level (NoiseLevel): Noise level of the dataset.
    """

    def __init__(self, noise_level: NoiseLevel = NoiseLevel.N0) -> None:
        """Initialize a MouseNuclei instance.

        Parameters
        ----------
        noise_level : NoiseLevel, optional
            Noise level of the dataset, by default NoiseLevel.N0
        """
        super().__init__(
            noise_level=noise_level,
            name=f"MouseNuclei_n{noise_level.value}",
            url=self._get_url(noise_level),
            file_name=f"MouseNuclei_n{noise_level.value}.zip",
            md5_hash=self._get_hash(noise_level),
            description="A dataset depicting diverse and non-uniformly "
            "clustered nuclei in the mouse skull, consisting of 908 training "
            "and 160 validation patches. The test set counts 67 additional images",
            license="CC BY-SA 4.0",
            citation="Buchholz, T.O., Prakash, M., Schmidt, D., Krull, A., Jug, "
            "F.: Denoiseg: joint denoising and segmentation. In: European "
            "Conference on Computer Vision (ECCV). pp. 324-337. Springer (2020) 8, 9",
            files={
                f"Mouse_n{noise_level.value}/train": ["train_data.npz"],
                f"Mouse_n{noise_level.value}/test": ["test_data.npz"],
            },
            size=self._get_size(noise_level),
        )

    @staticmethod
    def _get_url(noise: NoiseLevel) -> str:
        if noise == NoiseLevel.N0:
            return "https://zenodo.org/record/5157001/files/Mouse_n0.zip?download=1"
        elif noise == NoiseLevel.N10:
            return "https://zenodo.org/record/5157003/files/Mouse_n10.zip?download=1"
        else:
            return "https://zenodo.org/record/5157008/files/Mouse_n20.zip?download=1"

    @staticmethod
    def _get_hash(noise: NoiseLevel) -> str:
        if noise == NoiseLevel.N0:
            return "b747d013cba186a02c97937acef4b972"
        elif noise == NoiseLevel.N10:
            return "0b0776fa205057b49920b0ec3d1a5fc9"
        else:
            return "6e9d895ba3ac2c225883ed3ec94342f8"

    @staticmethod
    def _get_size(noise: NoiseLevel) -> float:
        if noise == NoiseLevel.N0:
            return 12.4
        elif noise == NoiseLevel.N10:
            return 161.0
        else:
            return 160.0



from pathlib import Path

from microscopy_portfolio import Portfolio

if __name__ == "__main__":
    # Create a portfolio object
    portfolio = Portfolio()

    # Path to the datasets list
    path_to_datasets = Path(".", "datasets", "datasets.json")

    # Export the portfolio to json
    portfolio.to_json(path_to_datasets)



import pytest
from microscopy_portfolio import Portfolio
from microscopy_portfolio.denoiseg_datasets import NoiseLevel, NoisyObject
from microscopy_portfolio.portfolio import PortfolioEntry

from .utils import (
    download_checker,
    portoflio_entry_checker,
    unique_md5_checker,
    unique_url_checker,
)

DATASETS = list(Portfolio().denoiseg)


@pytest.mark.dataset
@pytest.mark.parametrize("dataset", DATASETS)
def test_datasets(tmp_path, dataset: PortfolioEntry):
    """Test that all DenoiSeg datasets download properly.

    This test also checks the files and size.

    Parameters
    ----------
    tmp_path : Path
        Path to temporary directory.
    dataset : Dataset
        Dataset object.
    """
    download_checker(tmp_path, dataset)


def test_unique_hashes(portfolio: Portfolio):
    """Test that all DenoiSeg dataset hashes are unique.

    Parameters
    ----------
    portfolio : Portfolio
        Portfolio object.
    """
    unique_md5_checker(portfolio.denoiseg)


def test_unique_urls(portfolio: Portfolio):
    """Test that all DenoiSeg dataset URLs are unique.

    Parameters
    ----------
    portfolio : Portfolio
        Portfolio object.
    """
    unique_url_checker(portfolio.denoiseg)


def test_no_empty_dataset_entry(portfolio: Portfolio):
    """Test that no DenoiSeg dataset entry is empty.

    Parameters
    ----------
    portfolio : Portfolio
        Portfolio object.
    """
    for entry in portfolio.denoiseg:
        portoflio_entry_checker(entry)


@pytest.mark.parametrize("noise_level", [NoiseLevel.N0, NoiseLevel.N10, NoiseLevel.N20])
def test_noisy_dataset(noise_level):
    """Test that the NoisyDataset class works properly."""
    noisy_object = NoisyObject(noise_level=noise_level)
    assert noisy_object.noise_level == noise_level



import pytest
from microscopy_portfolio import Portfolio
from microscopy_portfolio.portfolio import IterablePortfolio


def list_iterable_portfolios():
    """List all iterable portfolios."""
    portfolio = Portfolio()

    list_iter_portfolios = []
    for attribute in vars(portfolio).values():
        if isinstance(attribute, IterablePortfolio):
            list_iter_portfolios.append(attribute)

    return list_iter_portfolios


ITERABLES = list_iterable_portfolios()


@pytest.mark.parametrize("iter_portfolio", ITERABLES)
def test_iterable_portfolios(iter_portfolio: IterablePortfolio):
    """Test that portfolios are iterable.

    Parameters
    ----------
    portfolio : Portfolio
        Portfolio to test.
    """
    entries = [entry.name for entry in iter_portfolio]
    assert entries == iter_portfolio.list_datasets()


@pytest.mark.parametrize("iter_portfolio", ITERABLES)
def test_iterable_portfolio_list_datasets(iter_portfolio: IterablePortfolio):
    """Test that the list_datasets method works on portfolios.

    Note: the list_datasets

    Parameters
    ----------
    portfolio : Portfolio
        Portfolio to test.
    """
    datasets = (
        iter_portfolio.list_datasets()
    )  # names of entries, rather than attributes
    assert len(datasets) > 0

    for entry in datasets:
        assert entry in iter_portfolio.as_dict().keys()


@pytest.mark.parametrize("iter_portfolio", ITERABLES)
def test_iterable_portfolio_as_dict(iter_portfolio: IterablePortfolio):
    # denoiseg
    iter_as_dict = iter_portfolio.as_dict()
    assert len(iter_as_dict) > 0

    # check entries
    for entry in iter_as_dict.values():
        assert "URL" in entry
        assert "Citation" in entry
        assert "Description" in entry
        assert "License" in entry
        assert "File size" in entry


def test_portfolio_as_dict(portfolio: Portfolio):
    """Test that the as_dict method works on portfolios.

    Parameters:
    -----------
    portfolio : Portfolio
        Portfolio to test.
    """
    portfolio_dict = portfolio.as_dict()
    assert len(portfolio_dict) > 0
    assert len(portfolio_dict) == len(ITERABLES)


def test_export_to_json(tmp_path, portfolio: Portfolio):
    """Test that the export Portfolio to json works.

    Parameters
    ----------
    tmp_path : Path
        Temporary path.
    portfolio : Portfolio
        Portfolio to test.
    """
    import json

    # export to json
    path_to_file = tmp_path / "portfolio.json"
    portfolio.to_json(path_to_file)

    # check that the file exists
    assert path_to_file.exists()

    # load json file
    with open(path_to_file) as f:
        data = json.load(f)

        for entry in ITERABLES:
            assert entry.name in data
            assert data[entry.name] == entry.as_dict()



"""Download and examine a dataset.

This script can be used to assert md5 and file size the first time
a dataset is added.
"""

import hashlib
import os
from pathlib import Path

from microscopy_portfolio import Portfolio

if __name__ == "__main__":
    # Create a portfolio object
    portfolio = Portfolio()

    # Download dataset
    root = Path("data")
    portfolio.denoising.Convallaria.download(root, check_md5=False)

    # Compute hash and size of downloaded file
    file_path = root / portfolio.denoising.Convallaria.file_name
    md5_hash = hashlib.md5(open(file_path, "rb").read()).hexdigest()
    size = os.path.getsize(file_path) / 1e6

    # Print results
    print(f"File: {file_path}")
    print(f"MD5 hash: {md5_hash}")
    print(f"Size: {size:.1f} MB")
